#include "mex.h"
#include <math.h>
#include <string.h>

#define MY_MAX(A, B) ((A)>(B)?(A):(B))
#define MY_MIN(A, B) ((A)<(B)?(A):(B))

void mexFunction (int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
    const mxArray* kernel_cacheMxArr;
    const mxArray* kernel_matMxArr;
    const mxArray* orderings_pt_idxsMxArr;
    const mxArray* box_constraintMxArr;
    const mxArray* doing_sparse_kernel_matMxArr;
    const mxArray* doing_sq_errorMxArr;

    mxArray* curr_sizeMxArr;

    double* kernel_mat;
    double* orderings_pt_idxs;

    double box_constraint;
    int doing_sparse_kernel_mat;
    int doing_sq_error;
    
    int n_pts, n_O;

    double* matrix;
    double* idx_to_row;
    double* row_to_idx;
    int max_size, curr_size;
   
    int idx_i, idx_j;
    int o_1, o_2; /* Copies of idx_i, idx_j */
    int row_i, row, col;

    double sum;

    if (nrhs != 1)
    {
        mexErrMsgIdAndTxt( "MATLAB:kernel_cache_add_row_pairwise:incorrectInputArguments",
                           "Cannot specify more or less than 1 inputs.\n");
    }
    
    kernel_cacheMxArr = mexGetVariablePtr("caller", "kernel_cache");
    if (kernel_cacheMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_add_row_pairwise:nullPointer",
                          "Variable 'kernel_cache' does not exist in workspace");
    }

    kernel_matMxArr = mexGetVariablePtr("caller", "kernel_mat");
    if (kernel_matMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_add_row_pairwise:nullPointer",
                          "Variable 'kernel_mat' does not exist in workspace");
    }

    orderings_pt_idxsMxArr = mexGetVariablePtr("caller", "orderings_pt_idxs");
    if (orderings_pt_idxsMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_add_row_pairwise:nullPointer",
                          "Variable 'orderings_pt_idxs' does not exist in workspace");
    }

    box_constraintMxArr = mexGetVariablePtr("caller", "box_constraint");
    if (box_constraintMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_add_row_pairwise:nullPointer",
                          "Variable 'box_constraint' does not exist in workspace");
    }

    doing_sparse_kernel_matMxArr = mexGetVariablePtr("caller", "doing_sparse_kernel_mat");    
    if (doing_sparse_kernel_matMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_add_row_pairwise:nullPointer",
                          "Variable 'doing_sparse_kernl_mat' does not exist in workspace");
    }

    doing_sq_errorMxArr = mexGetVariablePtr("caller", "doing_sq_error");    
    if (doing_sq_errorMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_add_row_pairwise:nullPointer",
                          "Variable 'doing_sq_error' does not exist in workspace");
    }

    /* Reading in globals */
    kernel_mat        = mxGetPr(kernel_matMxArr);
    orderings_pt_idxs = mxGetPr(orderings_pt_idxsMxArr);

    box_constraint          = mxGetScalar(box_constraintMxArr);
    doing_sparse_kernel_mat = (int) mxGetScalar(doing_sparse_kernel_matMxArr);
    doing_sq_error          = (int) mxGetScalar(doing_sq_errorMxArr);
    /*printf("doing_sparse_kernelmat: %d\n", doing_sparse_kernel_mat);*/
    /*printf("doing_sq_error: %d\n", doing_sq_error);*/

    if (mxIsStruct(kernel_cacheMxArr))
    { 
        
        max_size   = mxGetScalar(mxGetField(kernel_cacheMxArr, 0, "max_size"));
        curr_size  = mxGetScalar(mxGetField(kernel_cacheMxArr, 0, "curr_size"));
        idx_to_row = mxGetPr(mxGetField(kernel_cacheMxArr, 0, "idx_to_row"));
        row_to_idx = mxGetPr(mxGetField(kernel_cacheMxArr, 0, "row_to_idx"));
        matrix     = mxGetPr(mxGetField(kernel_cacheMxArr, 0, "matrix"));
        
        /*
        printf("Kernel cache max size: %d\n", max_size);
        printf("Kernel cache matrix(0, 0): %lf\n", matrix[0]);
        */
    }
    else
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_add_row_pairwise:notAStructure",
                          "Variable 'kernel_cache' is not a structure");

    }

    n_pts = mxGetM(kernel_matMxArr);
    n_O   = mxGetM(orderings_pt_idxsMxArr);
 
    /* Reading in arguments */
    idx_i = (int) mxGetScalar(prhs[0]);
    /*printf("idx_i: %d\n", idx_i);*/    

    row_i = (int) idx_to_row[idx_i-1];
    /*printf("row_i: %d\n", row_i);*/   

    if (row_i > 0)
    {
        return;
    }
    else if (curr_size >= max_size)
    {
        return;
    }    
    else
    {
        curr_size                = curr_size + 1;
        /*printf("curr_size: %d\n", curr_size);*/
        curr_sizeMxArr           = mxGetField(kernel_cacheMxArr, 0, "curr_size");
        *mxGetPr(curr_sizeMxArr) = (double) curr_size;
        /*printf("Kernel cache curr size: %lf\n", mxGetScalar(mxGetField(kernel_cacheMxArr, 0, "curr_size")));*/

        row_i               = curr_size;
        idx_to_row[idx_i-1] = row_i;
        row_to_idx[row_i-1] = idx_i;
        
        for (col=1; col <= curr_size; col++)
        {
            idx_j = row_to_idx[col-1];
            /*printf("idx_j: %d\n", idx_j);*/

            o_1 = idx_i;
            o_2 = idx_j;

            if (doing_sparse_kernel_mat)
            {
                int ik_max, ik_min;
                int il_max, il_min;
                int jk_max, jk_min;
                int jl_max, jl_min;

                ik_max = MY_MAX(orderings_pt_idxs[o_1-1], orderings_pt_idxs[o_2-1]);
                ik_min = MY_MIN(orderings_pt_idxs[o_1-1], orderings_pt_idxs[o_2-1]);
                il_max = MY_MAX(orderings_pt_idxs[o_1-1], orderings_pt_idxs[n_O + (o_2-1)]);
                il_min = MY_MIN(orderings_pt_idxs[o_1-1], orderings_pt_idxs[n_O + (o_2-1)]);
                jk_max = MY_MAX(orderings_pt_idxs[n_O + (o_1-1)], orderings_pt_idxs[o_2-1]);
                jk_min = MY_MIN(orderings_pt_idxs[n_O + (o_1-1)], orderings_pt_idxs[o_2-1]);
                jl_max = MY_MAX(orderings_pt_idxs[n_O + (o_1-1)], orderings_pt_idxs[n_O + (o_2-1)]);
                jl_min = MY_MIN(orderings_pt_idxs[n_O + (o_1-1)], orderings_pt_idxs[n_O + (o_2-1)]);

                sum  = 1.0;
                sum += kernel_mat[((ik_min-1)*n_pts)+(ik_max-1)];
                sum -= kernel_mat[((il_min-1)*n_pts)+(il_max-1)];
                sum -= kernel_mat[((jk_min-1)*n_pts)+(jk_max-1)];
                sum += kernel_mat[((jl_min-1)*n_pts)+(jl_max-1)];
            }
            else
            {
                int i, j, k, l;

                i = (int) orderings_pt_idxs[o_1-1];
                j = (int) orderings_pt_idxs[n_O + (o_1-1)];
                k = (int) orderings_pt_idxs[o_2-1];
                l = (int) orderings_pt_idxs[n_O + (o_2-1)];

                sum  = 1.0;
                sum += kernel_mat[((k-1)*n_pts)+(i-1)];
                sum -= kernel_mat[((l-1)*n_pts)+(i-1)];
                sum -= kernel_mat[((k-1)*n_pts)+(j-1)];
                sum += kernel_mat[((l-1)*n_pts)+(j-1)];
            }

            if (doing_sq_error)
            {
                if (o_1 == o_2)
                {
                    sum = sum + (1.0/box_constraint);
                }
            }

            matrix[((col-1)*max_size)+(row_i-1)] = sum; 
        }
    }

    return;
}
