#include "mex.h"

void mexFunction (int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
    const mxArray* kernel_cacheMxArr;

    mxArray* curr_sizeMxArr;

    double* matrix;
    double* idx_to_row;
    double* row_to_idx;
    int max_size, curr_size;

    int idx_i, idx_j;
    int row_i, col_i, row, col;
    int field_num;

    if (nrhs != 1)
    {
        mexErrMsgIdAndTxt( "MATLAB:kernel_cache_del_row:incorrectInputArguments",
                           "Cannot specify more or less than 1 inputs.\n");
    }
    
    kernel_cacheMxArr = mexGetVariablePtr("caller", "kernel_cache");
    if (kernel_cacheMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_del_row:nullPointer",
                          "Variable 'kernel_cache' does not exist in workspace");
    }

    /* Reading in globals */
    if (mxIsStruct(kernel_cacheMxArr))
    { 
        max_size   = mxGetScalar(mxGetField(kernel_cacheMxArr, 0, "max_size"));
        curr_size  = mxGetScalar(mxGetField(kernel_cacheMxArr, 0, "curr_size"));
        idx_to_row = mxGetPr(mxGetField(kernel_cacheMxArr, 0, "idx_to_row"));
        row_to_idx = mxGetPr(mxGetField(kernel_cacheMxArr, 0, "row_to_idx"));
        matrix     = mxGetPr(mxGetField(kernel_cacheMxArr, 0, "matrix"));
        /*
        printf("Kernel cache max size: %d\n", max_size);
        printf("Kernel cache matrix(0, 0): %lf\n", matrix[0]);
        */
    }
    else
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_cache_del_row:notAStructure",
                          "Variable 'kernel_cache' is not a structure");

    }
 
    /* Reading in arguments */
    idx_i = (int) mxGetScalar(prhs[0]);
    /*printf("idx_i: %d\n", idx_i);*/    

    row_i = (int) idx_to_row[idx_i-1];
    /*printf("row_i: %d\n", row_i);*/   

    if (row_i <= 0)
    {
        return;
    }
    else if (curr_size < max_size)
    {
        return;
    }
    else
    {
        curr_size                = curr_size - 1;
        /*printf("curr_size: %d\n", curr_size);*/
        curr_sizeMxArr           = mxGetField(kernel_cacheMxArr, 0, "curr_size");
        *mxGetPr(curr_sizeMxArr) = (double) curr_size;
        /*printf("Kernel cache curr size: %lf\n", mxGetScalar(mxGetField(kernel_cacheMxArr, 0, "curr_size")));*/

        col_i               = row_i;
        idx_to_row[idx_i-1] = 0;
        
        if(row_i == max_size)
        {
            for (col=1; col <= max_size; col++)
            {
                matrix[(col-1)*max_size+(row_i-1)] = 0;
            }
            /*
            for (row=1; row <= max_size; row++)
            {
                matrix[(col_i-1)*max_size+(row-1)] = 0;
            }
            */
            row_to_idx[row_i-1] = 0;
        }
        else
        {
            for (row=row_i; row <= curr_size; row++)
            {
                /*for (col=1; col <= max_size; col++)*/
                for (col=0; col <= curr_size; col++)
                {
                    /*matrix[(col-1)*max_size+(row-1)] = matrix[(col-1)*max_size+row];*/
                    matrix[(col)*max_size+(row-1)] = matrix[(col)*max_size+row];
                    
                }
                row_to_idx[row-1] = row_to_idx[row];
                idx_to_row[((int)row_to_idx[row])-1] = (double) row;
            }
            for (col=col_i; col <= curr_size; col++)
            {
                for (row=row_i; row <= curr_size; row++)
                {
                    matrix[(col-1)*max_size+(row-1)] = matrix[(col)*max_size+(row-1)]; 
                }
            }
            /*for (col=1,row=1; col <= max_size; col++,row++)*/
            for (col=0,row=0; col <= curr_size; col++,row++)
            {
                /*
                matrix[(col-1)*max_size+(curr_size)] = 0;
                matrix[(curr_size)*max_size+(row-1)] = 0;
                */
                matrix[(col)*max_size+(curr_size)] = 0;
                matrix[(curr_size)*max_size+(row)] = 0;
            }
            row_to_idx[curr_size] = 0;
        }
    }

    return;
}
