#include "mex.h"

#define MY_MAX(A, B) ((A)>(B)?(A):(B))
#define MY_MIN(A, B) ((A)<(B)?(A):(B))

void mexFunction (int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
    const mxArray* kernel_matMxArr;
    const mxArray* orderings_pt_idxsMxArr;
    const mxArray* box_constraintMxArr;
    const mxArray* doing_sparse_kernel_matMxArr;
    const mxArray* doing_sq_errorMxArr;

    double* kernel_mat;
    double* orderings_pt_idxs;

    double box_constraint;
    int doing_sparse_kernel_mat;
    int doing_sq_error;
    
    int n_pts, n_O;
    int o_1, o_2;

    double retVal;

    if (nrhs != 2)
    {
        mexErrMsgIdAndTxt( "MATLAB:kernel_mat_pairwise_fn:incorrectInputArguments",
                           "Cannot specify more or less than 2 inputs.\n");
    }
    
    if (nlhs != 1)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_mat_pairwise_fn:incorrectOutputArguments", 
                          "Cannot specify more or less than 1 input.\n");
    }    

    kernel_matMxArr = mexGetVariablePtr("caller", "kernel_mat");
    if (kernel_matMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_mat_pairwise_fn:nullPointer",
                          "Variable 'kernel_mat' does not exist in workspace");
    }

    orderings_pt_idxsMxArr = mexGetVariablePtr("caller", "orderings_pt_idxs");
    if (orderings_pt_idxsMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_mat_pairwise_fn:nullPointer",
                          "Variable 'orderings_pt_idxs' does not exist in workspace");
    }

    box_constraintMxArr = mexGetVariablePtr("caller", "box_constraint");
    if (box_constraintMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_mat_pairwise_fn:nullPointer",
                          "Variable 'box_constraint' does not exist in workspace");
    }

    doing_sparse_kernel_matMxArr = mexGetVariablePtr("caller", "doing_sparse_kernel_mat");    
    if (doing_sparse_kernel_matMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_mat_pairwise_fn:nullPointer",
                          "Variable 'doing_sparse_kernl_mat' does not exist in workspace");
    }

    doing_sq_errorMxArr = mexGetVariablePtr("caller", "doing_sq_error");    
    if (doing_sq_errorMxArr == NULL)
    {
        mexErrMsgIdAndTxt("MATLAB:kernel_mat_pairwise_fn:nullPointer",
                          "Variable 'doing_sq_error' does not exist in workspace");
    }

    /* Reading in globals */
    kernel_mat        = mxGetPr(kernel_matMxArr);
    orderings_pt_idxs = mxGetPr(orderings_pt_idxsMxArr);

    box_constraint          = mxGetScalar(box_constraintMxArr);
    doing_sparse_kernel_mat = (int) mxGetScalar(doing_sparse_kernel_matMxArr);
    doing_sq_error          = (int) mxGetScalar(doing_sq_errorMxArr);
    /*printf("doing_sparse_kernelmat: %d\n", doing_sparse_kernel_mat);*/
    /*printf("doing_sq_error: %d\n", doing_sq_error);*/
     
    /* Reading in arguments */
    o_1 = (int) mxGetScalar(prhs[0]);
    o_2 = (int) mxGetScalar(prhs[1]);
    /*printf("o_1: %d, o_2: %d\n", o_1, o_2);*/

    n_pts = mxGetM(kernel_matMxArr);
    n_O   = mxGetM(orderings_pt_idxsMxArr);

    if (doing_sparse_kernel_mat)
    {
        int ik_max, ik_min;
        int il_max, il_min;
        int jk_max, jk_min;
        int jl_max, jl_min;

        ik_max = MY_MAX(orderings_pt_idxs[o_1-1], orderings_pt_idxs[o_2-1]);
        ik_min = MY_MIN(orderings_pt_idxs[o_1-1], orderings_pt_idxs[o_2-1]);
        il_max = MY_MAX(orderings_pt_idxs[o_1-1], orderings_pt_idxs[n_O + (o_2-1)]);
        il_min = MY_MIN(orderings_pt_idxs[o_1-1], orderings_pt_idxs[n_O + (o_2-1)]);
        jk_max = MY_MAX(orderings_pt_idxs[n_O + (o_1-1)], orderings_pt_idxs[o_2-1]);
        jk_min = MY_MIN(orderings_pt_idxs[n_O + (o_1-1)], orderings_pt_idxs[o_2-1]);
        jl_max = MY_MAX(orderings_pt_idxs[n_O + (o_1-1)], orderings_pt_idxs[n_O + (o_2-1)]);
        jl_min = MY_MIN(orderings_pt_idxs[n_O + (o_1-1)], orderings_pt_idxs[n_O + (o_2-1)]);

        retVal  = 1.0;
        retVal += kernel_mat[((ik_min-1)*n_pts)+(ik_max-1)];
        retVal -= kernel_mat[((il_min-1)*n_pts)+(il_max-1)];
        retVal -= kernel_mat[((jk_min-1)*n_pts)+(jk_max-1)];
        retVal += kernel_mat[((jl_min-1)*n_pts)+(jl_max-1)];
    }
    else
    {
        int i, j, k, l;

        i = (int) orderings_pt_idxs[o_1-1];
        j = (int) orderings_pt_idxs[n_O + (o_1-1)];
        k = (int) orderings_pt_idxs[o_2-1];
        l = (int) orderings_pt_idxs[n_O + (o_2-1)];

        retVal  = 1.0;
        retVal += kernel_mat[((k-1)*n_pts)+(i-1)];
        retVal -= kernel_mat[((l-1)*n_pts)+(i-1)];
        retVal -= kernel_mat[((k-1)*n_pts)+(j-1)];
        retVal += kernel_mat[((l-1)*n_pts)+(j-1)];
    }

    if (doing_sq_error)
    {
        if (o_1 == o_2)
        {
            retVal = retVal + (1.0/box_constraint);
        }
    }

    /* Return result */
    plhs[0] = mxCreateDoubleScalar(retVal);

    return;
}
